'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
var BOUNDS_FORMAT = ['left', 'top', 'right', 'bottom'];

var MIRROR_LR = {
  center: 'center',
  left: 'right',
  right: 'left'
};

var MIRROR_TB = {
  middle: 'middle',
  top: 'bottom',
  bottom: 'top'
};

/**
 * This function returns a rect 2px x 2px with the origin based at (1,1), which allows for evenly centered attachment points.
 * As a result of the 2x2 size, positioning issues caused by coordinate rounding are eliminated.
 *
 * @ param {object} coords - THe x and y client coordinates.
 */
var getRectFromCoords = function getRectFromCoords(coords) {
  return {
    top: coords.y - 1,
    right: coords.x + 1,
    bottom: coords.y + 1,
    left: coords.x - 1,
    width: 2,
    height: 2
  };
};

/**
 * This function returns the size of an element and its position relative to the viewport. It takes into account parent <frame> offsets
 * if the element lies within a nested document (<frame> or <iframe>-like).
 *
 * @ param {htmlELement} node - The html reference to use to retrieve it's bounding rect values.
 */
var getActualBoundingClientRect = function getActualBoundingClientRect(node) {
  var clientRect = node.getBoundingClientRect();
  var rect = {
    top: clientRect.top,
    right: clientRect.right,
    bottom: clientRect.bottom,
    left: clientRect.left,
    width: clientRect.width,
    height: clientRect.height
  };

  if (node.ownerDocument !== document) {
    // eslint-disable-next-line prefer-destructuring
    var frameElement = node.ownerDocument.defaultView.frameElement;
    if (frameElement) {
      var frameRect = getActualBoundingClientRect(frameElement);
      rect.top += frameRect.top;
      rect.bottom += frameRect.top;
      rect.left += frameRect.left;
      rect.right += frameRect.left;
    }
  }

  return rect;
};

/**
 * This function returns the list of parent elements capable of scrolling until a fixed element or document is found.
 *
 * @ param {htmlELement} element - The html reference to use to retrieve the parent nodes.
 */
var getScrollParents = function getScrollParents(element) {
  var computedStyle = getComputedStyle(element) || {};
  var position = computedStyle.position;

  var parents = [];

  if (position === 'fixed') {
    return [element];
  }

  var parent = element.parentNode;
  while (parent && parent.nodeType === 1) {
    var style = void 0;
    try {
      style = getComputedStyle(parent);
    } catch (err) {
      style = null;
    }

    if (typeof style === 'undefined' || style === null) {
      parents.push(parent);
      return parents;
    }

    var _style = style,
        overflow = _style.overflow,
        overflowX = _style.overflowX,
        overflowY = _style.overflowY;

    if (/(auto|scroll|overlay)/.test(overflow + overflowY + overflowX)) {
      if (position !== 'absolute' || ['relative', 'absolute', 'fixed', 'static'].indexOf(style.position) >= 0) {
        parents.push(parent);
      }
    }
    parent = parent.parentNode;
  }

  parents.push(element.ownerDocument.body);

  // If the node is within a frame, account for the parent window scroll
  if (element.ownerDocument !== document) {
    parents.push(element.ownerDocument.defaultView);
  }

  return parents;
};

/**
 * This function returns the bounds of an element.
 *
 * @ param {htmlELement} element - The html reference to use to retrieve it's bounds.
 */
var getBounds = function getBounds(element) {
  var doc = document;
  var currentElement = document.documentElement;

  if (element !== document) {
    doc = element.ownerDocument;
    currentElement = element;
  }

  var docEl = doc.documentElement;
  var box = getActualBoundingClientRect(currentElement);

  if (typeof box.width === 'undefined') {
    box.width = document.body.scrollWidth - box.left - box.right;
  }

  if (typeof box.height === 'undefined') {
    box.height = document.body.scrollHeight - box.top - box.bottom;
  }

  box.top -= docEl.clientTop;
  box.left -= docEl.clientLeft;
  box.right = doc.body.clientWidth - box.width - box.left;
  box.bottom = doc.body.clientHeight - box.height - box.top;

  // round origin
  box.top = Math.round(box.top);
  box.right = Math.round(box.right);
  box.bottom = Math.round(box.bottom);
  box.left = Math.round(box.left);
  box.height = Math.round(box.height);
  box.width = Math.round(box.width);

  return box;
};

/**
 * This function returns the size of an element. It takes into account scroll offsets and border widths applied.
 *
 * @ param {htmlELement/string} boundingElement - The value used to determine the size of bounding element.
 */
var getBoundingRect = function getBoundingRect(boundingElement) {
  if (boundingElement === 'window') {
    return {
      top: 0,
      // eslint-disable-next-line no-restricted-globals
      bottom: innerHeight,
      left: 0,
      // eslint-disable-next-line no-restricted-globals
      right: innerWidth
    };
  }

  var bounds = getBounds(boundingElement);
  var style = getComputedStyle(boundingElement);
  var rect = {
    top: bounds.top,
    bottom: bounds.top + bounds.height,
    left: bounds.left,
    right: bounds.left + bounds.width
  };

  // Account for any parent Frames scroll offset
  if (boundingElement.ownerDocument !== document) {
    var parentView = boundingElement.ownerDocument.defaultView;

    rect.left += parentView.pageXOffset;
    rect.top += parentView.pageYOffset;
    rect.right += parentView.pageXOffset;
    rect.bottom += parentView.pageYOffset;
  }

  BOUNDS_FORMAT.forEach(function (side) {
    var subSide = side[0].toUpperCase() + side.substr(1);
    if (subSide === 'Top' || subSide === 'Left') {
      rect[side] += parseFloat(style['border' + subSide + 'Width']);
    } else {
      rect[side] -= parseFloat(style['border' + subSide + 'Width']);
    }
  });

  // round origin
  rect.top = Math.round(rect.top);
  rect.right = Math.round(rect.right);
  rect.bottom = Math.round(rect.bottom);
  rect.left = Math.round(rect.left);

  return rect;
};

/**
 * This object contains the attachment values for the target when using coordinate based positioning.
 */
var coordinateAttachment = { vertical: 'middle', horizontal: 'center' };

/**
 * This function returns the attachment object, adjusted for RTL conversion.
 *
 * @ param {Object} attachment - The vertical and horizonal hookshot attachments.
 * @ param {bool} isRTL - Whether or not the page is using RTL.
 */
var getDirectionalAttachment = function getDirectionalAttachment(attachment, isRTL) {
  if (attachment.horizontal === 'start') {
    return { vertical: attachment.vertical, horizontal: isRTL ? 'right' : 'left' };
  }if (attachment.horizontal === 'end') {
    return { vertical: attachment.vertical, horizontal: isRTL ? 'left' : 'right' };
  }
  return attachment;
};

/**
 * This function returns the offset object, adjusted for RTL conversion.
 *
 * @ param {Object} offset - The vertical and horizonal offset values.
 * @ param {bool} isRTL - Whether or not the page is using RTL.
 */
var getDirectionalOffset = function getDirectionalOffset(offset, isRTL) {
  if (isRTL) {
    return { vertical: offset.vertical, horizontal: -offset.horizontal };
  }
  return offset;
};

/**
 * This function returns an attachment object containing the mirrored attachment values.
 *
 * @ param {Object} attachment - The vertical and horizonal hookshot attachments.
 */
var mirrorAttachment = function mirrorAttachment(attachment) {
  if (attachment.vertical !== 'middle') {
    return { vertical: MIRROR_TB[attachment.vertical], horizontal: attachment.horizontal };
  }
  return { vertical: attachment.vertical, horizontal: MIRROR_LR[attachment.horizontal] };
};

/**
 * This function returns an offset object containing the mirrored offset values.
 *
 * @ param {Object} offset - The vertical and horizonal offset values.
 * @ param {Object} attachment - The vertical and horizonal hookshot attachments.
 */
var mirrorOffset = function mirrorOffset(offset, attachment) {
  if (attachment.vertical !== 'middle') {
    return { vertical: -offset.vertical, horizontal: offset.horizontal };
  }
  return { vertical: offset.vertical, horizontal: -offset.horizontal };
};

/**
 * This function returns an attachment object containing the rotated attachment values given an angle value.
 *
 * @ param {Object} attachment - The vertical and horizonal hookshot attachments.
 * @ param {string} angle - Representation of the angle to rotate the attachment to.
 */
var rotateContentAttachment = function rotateContentAttachment(attachment, angle) {
  var rAttachment = {};

  if (attachment.vertical === 'middle') {
    if (angle === '90') {
      rAttachment.vertical = attachment.horizontal === 'left' ? 'bottom' : 'top';
    } else if (angle === '-90') {
      rAttachment.vertical = attachment.horizontal === 'left' ? 'top' : 'bottom';
    }
    rAttachment.horizontal = 'center';
  } else {
    if (angle === '90') {
      rAttachment.horizontal = attachment.vertical === 'top' ? 'left' : 'right';
    } else if (angle === '-90') {
      rAttachment.horizontal = attachment.vertical === 'top' ? 'right' : 'left';
    }
    rAttachment.vertical = 'middle';
  }

  return rAttachment;
};

/**
 * This function returns whether or not there is available space to place the content given the provided bounds and positions of
 * the content and bounding elements.
 *
 * @ param {Object} positions - The positional data of the content and target to evaluate.
 * @ param {Object} boundingRect - The bounding rectangle.
 */
var isValidPositions = function isValidPositions(positions, boundingRect) {
  var content = positions.content;
  var attachment = content.attachment;

  var contentRect = content.rect;

  if (attachment.vertical === 'middle') {
    if (attachment.horizontal === 'right') {
      return content.x >= boundingRect.left;
    }if (attachment.horizontal === 'left') {
      return content.x + contentRect.width <= boundingRect.right;
    }
    return true;
  }if (attachment.vertical === 'top') {
    return content.y + contentRect.height <= boundingRect.bottom;
  }

  return content.y >= boundingRect.top;
};

/**
 * This function returns the screen coordinates, attachment values and offset values for the target.
 *
 * @ param {Object} target - The target x/y coords, rectangle, offset, and attachment.
 */
var getTargetPosition = function getTargetPosition(target) {
  var targetCoords = {};
  var targetTop = target.rect.top;
  var targetHeight = target.rect.height;
  var targetLeft = target.rect.left;
  var targetWidth = target.rect.width;

  if (target.attachment.vertical === 'middle') {
    targetCoords.y = targetTop + targetHeight / 2;
  } else if (target.attachment.vertical === 'bottom') {
    targetCoords.y = targetTop + targetHeight;
  } else {
    targetCoords.y = targetTop;
  }

  if (target.attachment.horizontal === 'center') {
    targetCoords.x = targetLeft + targetWidth / 2;
  } else if (target.attachment.horizontal === 'right') {
    targetCoords.x = targetLeft + targetWidth;
  } else {
    targetCoords.x = targetLeft;
  }

  return {
    x: targetCoords.x + target.offset.horizontal,
    y: targetCoords.y + target.offset.vertical,
    attachment: target.attachment,
    offset: target.offset,
    rect: target.rect
  };
};

/**
 * This function returns screen coordinates, attchment values and offset values of the target if mirrored.
 *
 * @ param {Object} target - The target x/y coords, rectangle, offset, and attachment.
 * @ param {Object} contentAttachment - The vertical and horizonal hookshot attachments of the content.
 */
var mirrorTargetPosition = function mirrorTargetPosition(target, contentAttachment) {
  var mirroredTarget = {
    rect: target.rect,
    attachment: target.attachment,
    offset: mirrorOffset(target.offset, target.attachment)
  };

  if (contentAttachment.vertical === target.attachment.vertical || contentAttachment.vertical !== 'middle' && target.attachment.vertical !== 'middle') {
    mirroredTarget.attachment = mirrorAttachment(target.attachment);
  }

  return getTargetPosition(mirroredTarget);
};

/**
 * This function returns the relative positional data of the content and target. Positional data includes
 * the screen coordinates, attchment values and offset values.
 *
 * @ param {Object} content - The content x/y coords, rectangle, offset, and attachment.
 * @ param {Object} target - The target x/y coords, rectangle, offset, and attachment.
 * @ param {number} margin - The px value of the attachmentMargin.
 */
var getRelativePositions = function getRelativePositions(content, target, margin) {
  var contentCoords = {};
  var contentHeight = content.rect.height;
  var contentWidth = content.rect.width;

  if (content.attachment.vertical === 'middle') {
    if (content.attachment.horizontal === 'center') {
      contentCoords.x = target.x - contentWidth / 2;
    } else if (content.attachment.horizontal === 'right') {
      contentCoords.x = target.x - contentWidth - margin;
    } else {
      contentCoords.x = target.x + margin;
    }

    contentCoords.y = target.y - contentHeight / 2;
  } else {
    if (content.attachment.horizontal === 'center') {
      contentCoords.x = target.x - contentWidth / 2;
    } else if (content.attachment.horizontal === 'right') {
      contentCoords.x = target.x - contentWidth;
    } else {
      contentCoords.x = target.x;
    }

    if (content.attachment.vertical === 'bottom') {
      contentCoords.y = target.y - contentHeight - margin;
    } else {
      contentCoords.y = target.y + margin;
    }
  }

  return {
    content: {
      x: contentCoords.x + content.offset.horizontal,
      y: contentCoords.y + content.offset.vertical,
      attachment: content.attachment,
      offset: content.offset,
      rect: content.rect
    },
    target: target
  };
};

/**
 * This function attempts to flip the positions given the attachmnt behavior and returns the corresponding positional data.
 * Positional data includes the screen coordinates, attachment values and offset values.
 *
 * @ param {Object} positions - The relative positional data of the content and target to evaluate.
 * @ param {number} margin - The px value of the attachmentMargin.
 */
var getMirroredPositions = function getMirroredPositions(positions, margin) {
  var newContent = {
    offset: mirrorOffset(positions.content.offset, positions.content.attachment),
    attachment: mirrorAttachment(positions.content.attachment),
    rect: positions.content.rect
  };

  return getRelativePositions(newContent, mirrorTargetPosition(positions.target, positions.content.attachment), margin);
};

/**
 * This function attempts to rotate the positions given the attachmnt behavior and returns the corresponding positional data.
 * Positional data includes the screen coordinates, attachment values and offset values.
 *
 * @ param {Object} positions - The relative positional data of the content and target to evaluate.
 * @ param {number} margin - The px value of the attachmentMargin.
 * @ param {string} rotation - The rotational angle to attach.
 */
var getRotatedPositions = function getRotatedPositions(positions, margin, rotation) {
  var newContent = {
    offset: { vertical: 0, horizontal: 0 },
    attachment: rotateContentAttachment(positions.content.attachment, rotation),
    rect: positions.content.rect
  };
  var newTarget = {
    offset: { vertical: 0, horizontal: 0 },
    attachment: mirrorAttachment(newContent.attachment),
    rect: positions.target.rect
  };

  return getRelativePositions(newContent, getTargetPosition(newTarget), margin);
};

/**
 * This function attempts to rotate the content element given the attachmnt behavior and returns the corresponding positional data.
 * Positional data includes the screen coordinates, attchment values and offset values.
 *
 * @ param {Object} positions - The relative positional data of the content and target to evaluate.
 * @ param {Object} boundingRect - The bounding rectangle.
 * @ param {number} margin - The px value of the attachmentMargin.
 * @ param {string} behavior - The attachemnt behavior which indicates the available content rotations.
 */
var getAlternatePositions = function getAlternatePositions(positions, boundingRect, margin, behavior) {
  if (behavior === 'auto' || behavior === 'flip') {
    // Attempt to flip content 180 degrees
    var newPositions = getMirroredPositions(positions, margin);
    if (isValidPositions(newPositions, boundingRect)) {
      return newPositions; // 180 degree rotation is valid
    }

    if (behavior === 'auto') {
      // Attempt to flip content 90 degrees
      newPositions = getRotatedPositions(positions, margin, '90');
      if (isValidPositions(newPositions, boundingRect)) {
        return newPositions; // 90 degree rotation is valid
      }

      // Attempt to flip content -90 degrees
      newPositions = getRotatedPositions(positions, margin, '-90');
      if (isValidPositions(newPositions, boundingRect)) {
        return newPositions; // -90 degree rotation is valid
      }
    }
  }

  // No alternate positions found, return original value
  return positions;
};

/**
 * This function adjusts the contents positional data to ensure it fits for bounding rectangle. Returns weather or
 * not the content was bounded and the positional data of the content and target to evaluate. Positional data includes
 * the screen coordinates, attchment values and offset values.
 *
 * @ param {Object} positions - The positional data of the content and target to evaluate.
 * @ param {Object} boundingRect - The bounding rectangle.
 */
var getBoundedPositions = function getBoundedPositions(positions, boundingRect) {
  var contentCoords = {};
  var contentRect = positions.content.rect;
  var contentX = positions.content.x;
  var contentY = positions.content.y;

  // Checks Horizontal Bounds
  if (boundingRect.left >= contentX) {
    contentCoords.x = boundingRect.left;
  } else if (boundingRect.right <= contentX + contentRect.width) {
    contentCoords.x = boundingRect.right - contentRect.width;
  } else {
    contentCoords.x = contentX;
  }

  // Checks Vertical Bounds
  if (boundingRect.top >= contentY) {
    contentCoords.y = boundingRect.top;
  } else if (boundingRect.bottom <= contentY + contentRect.height) {
    contentCoords.y = boundingRect.bottom - contentRect.height;
  } else {
    contentCoords.y = contentY;
  }

  return {
    content: {
      x: contentCoords.x,
      y: contentCoords.y,
      attachment: positions.content.cAttachment,
      offset: positions.content.cOffset,
      rect: contentRect
    },
    target: positions.target
  };
};

/**
 * The function returns the style and positional data the hookshot content and target. Positional data includes
 * the screen coordinates, attchment values and offset values.
 *
 * @ param {Object} boundingRect - The bounding rectangle.
 * @ param {Object} content - The content x/y coords, rectangle, offset, and attachment.
 * @ param {Object} target - The target x/y coords, rectangle, offset, and attachment.
 * @ param {number} margin - The px value of the attachmentMargin.
 * @ param {string} behavior - The attachemnt behavior which indicates the available content rotations.
 */
var positionStyleFromBounds = function positionStyleFromBounds(boundingRect, content, target, margin, behavior) {
  var targetPosition = getTargetPosition(target);

  // Get relative content and target positions
  var positions = getRelativePositions(content, targetPosition, margin);

  if (boundingRect) {
    if (!isValidPositions(positions, boundingRect)) {
      // Try to find valid alternative positions
      positions = getAlternatePositions(positions, boundingRect, margin, behavior);
    }

    positions = getBoundedPositions(positions, boundingRect);
  }

  return {
    style: {
      position: 'absolute',
      // eslint-disable-next-line no-restricted-globals
      left: Math.round(positions.content.x + pageXOffset) + 'px',
      // eslint-disable-next-line no-restricted-globals
      top: Math.round(positions.content.y + pageYOffset) + 'px'
    },
    positions: positions
  };
};

exports.default = {
  getScrollParents: getScrollParents,
  getRectFromCoords: getRectFromCoords,
  getBounds: getBounds,
  getBoundingRect: getBoundingRect,
  mirrorAttachment: mirrorAttachment,
  coordinateAttachment: coordinateAttachment,
  getDirectionalAttachment: getDirectionalAttachment,
  getDirectionalOffset: getDirectionalOffset,
  positionStyleFromBounds: positionStyleFromBounds
};